package frfs2;

//TODO - handle missing values?

import java.util.*;

import set.FuzzySet;
import set.Subset;

public final class RSys2 {
	int attr; // number of attributes in the current reduct
	int n_attributes; // the number of attributes in the system
	int n_objects; // the number of objects in the system
	ArrayList<BitSet> decision; // the indiscernible eqv classes of the decision
	// attr
	static Matrix fm;
	Subset red,Reduct;
	boolean verbose = false, end;
	FuzzySet[] decisionSets;
	Relation[] ret;

	boolean noTransitive = false;
	boolean roughEntropy = false;

	double gamma = 0;
	Relation[] relations;
	private double decCardinality = 0;
	private double[] decisionLength;

	// for boundary region calculation
	double[] lowers, uppers;

	// for optimization
	int[] indexes;
	double[] vals;
	int[] decIndexes;
	double[] decVals;
	Relation temprel;
	Relation best, current;
	boolean[] noChange;

	// for dataset statistics (per attribute)
	double[] attrDifference;
	double[] attrSTD;
	double[] mean;
	double[] max;
	double[] min;

	// double equivalent
	double n_objects_d;

	// cardinality of each fuzzy eq class
	double cards[];

	// normalising factor
	double c_divisor = 1; // may need to change this

	// similarity measures
	public int SIMILARITY = 0;
	public static final int GAUSSIAN = 0;
	public static final int TRIANGULAR = 1;
	public static final int TRIANGULAR2 = 2;
	public static final int EQUIVALENCE = 3;
	public static final int TRIANGULAR3 = 4;
	public static final int CORNELISA = 5;

	// VQRS-specific parameters/values
	double alpha_l, beta_l;
	double alpha_u, beta_u;
	double denomVal;
	double dependencyFull=1;

	// Fuzzy entropy-related values
	public final static double LN2 = 0.69314718055994531;
	double prevEntropy = 0;
	double entropy;

	public RSys2(Matrix ftbl, int objs, boolean v, int sim, boolean nt,
			double al, double be, double al2, double be2) {
		fm = ftbl;
		verbose = v;
		n_objects = objs;
		n_objects_d = n_objects;
		cards = new double[n_objects];
		n_attributes = ftbl.numCols();// ftbl.cols();
		SIMILARITY = sim;
		noTransitive = nt;
		decision = new ArrayList<BitSet>();

		// alpha and beta for VQRS lower approx
		alpha_u = al;
		beta_u = be;

		// alpha and beta for VQRS upper approx
		alpha_l = al2;
		beta_l = be2;

		current = new Relation(n_objects);
		temprel = new Relation(n_objects);

		setAttributeStats();
		if (frfs.OPTIMISE < 2)
			constructSimilarity(n_attributes); // need similarity for decision

		// work out values associated with the normalisation factor in
		// entropy/boundary calculations
		if (frfs.command == frfs.BOUNDARY
				|| frfs.command == frfs.ENTROPY_REDUCT
				|| frfs.command == frfs.INTERVAL_BOUNDARY
				|| frfs.command == frfs.BOUNDARY_ENTROPY
				|| frfs.command == frfs.GAIN_RATIO) {

			decisionLength = new double[n_objects];

			for (int x = 0; x < n_objects; x++) {
				for (int y = 0; y < n_objects; y++) {
					double decSim = getDecisionSimilarity(null, x, y);
					decisionLength[x] += decSim;
				}

				decCardinality += 1 / decisionLength[x];
			}
		}
		decIndexes = new int[n_objects];

		// for optimizations - check to see if decision classes are the same
		if (frfs.command != frfs.NN || frfs.command != frfs.FNN
				|| frfs.command != frfs.FRNN || frfs.command != frfs.VQNN) {
			for (int i = 0; i < n_objects; i++) {
				boolean same = false;

				for (int j = i + 1; j < n_objects; j++) {

					for (int d = 0; d < n_objects; d++) {
						double decSim1 = getDecisionSimilarity(null, i, d);
						double decSim2 = getDecisionSimilarity(null, j, d);

						if (decSim1!=decSim2) {same=false; break;}
						else same=true;

					}
					if (same) {
						decIndexes[j] = i + 1;
						break;
					} // i+1 as default is 0
				}

			}
		}
		// construct fuzzy sets for the decision feature if it is fuzzy
		if (frfs.fuzzyDecision) {
			decisionSets = new FuzzySet[5];
			float threshold = 0.7f;
			float mean1 = (float) mean[n_attributes - 1];
			float std = (float) attrSTD[n_attributes - 1];
			float point1 = mean1 - threshold * std;
			float point2 = mean1 + threshold * std;

			decisionSets[0] = new FuzzySet("Low", n_attributes - 1,
					(mean1 - 2 * std), (mean1 - std), 0.0f, 0.0f, FuzzySet.L_SHOULDERED); 
			decisionSets[1] = new FuzzySet("Low-Med", n_attributes - 1,
					(mean1 - 2 * std), (mean1 - std), point1, 0.0f, FuzzySet.TRIANGULAR);
			decisionSets[2] = new FuzzySet("Med", n_attributes - 1,
					(mean1 - std), point1, point2, (mean1 + std),
					FuzzySet.TRAPEZOIDAL);
			decisionSets[3] = new FuzzySet("Med-High", n_attributes - 1,
					point2, (mean1 + std), (mean1 + 2 * std), 0.0f,
					FuzzySet.TRIANGULAR);
			decisionSets[4] = new FuzzySet("High", n_attributes - 1,
					(mean1 + std), (mean1 + 2 * std), 0.0f, 0.0f,
					FuzzySet.R_SHOULDERED);
		}
		setConsistency();
	}

	// construct fuzzy similarity relations
	private final void constructSimilarity(int attributes) {
		relations = new Relation[attributes];
		double mainValue, thisValue;

		for (int a = 0; a < attributes; a++) {
			relations[a] = new Relation(n_objects);
			for (int mainObject = 0; mainObject < n_objects; mainObject++) {

				mainValue = fm.getCell(mainObject, a);
				relations[a].setCell(mainObject, mainObject, 1);
				for (int o = mainObject + 1; o < n_objects; o++) {
					thisValue = fuzzySimilarity(a, mainValue, fm.getCell(o, a));
					relations[a].setCell(mainObject, o, thisValue);
					relations[a].setCell(o, mainObject, thisValue);
				}
			}
		}

		noChange = new boolean[n_attributes - 1];

		if (!noTransitive) {
			System.err.print("Computing transitive closure");
			// should be repeated n_objects-1 times (or until no change)

			for (int i = 0; i < n_objects - 1; i++) {
				if (i % (int) (n_objects_d / 4) == 0)
					System.err.print(".");
				transitiveClosure();

			}
			System.err.println("done");
		}
	}

	// compute the transitive closure of the fuzzy similarity relation
	// for all conditional attributes
	private final void transitiveClosure() {
		ret = new Relation[n_attributes - 1];
		// System.arraycopy(relations,0,ret,0,relations.length);

		for (int a = 0; a < n_attributes - 1; a++) {
			ret[a] = new Relation(n_objects);
			for (int i = 0; i < n_objects; i++) {
				for (int j = 0; j < n_objects; j++) {
					ret[a].setCell(i, j, relations[a].getCell(i, j));
				}
			}
		}

		for (int a = 0; a < n_attributes - 1; a++) {
			if (!noChange[a]) {
				noChange[a] = true;
				for (int x = 0; x < n_objects; x++) {
					for (int y = x; y < n_objects; y++) {
						double v = closure(a, x, y);
						if (ret[a].getCell(x, y) != v)
							noChange[a] = false;
						ret[a].setCell(x, y, v);
						ret[a].setCell(y, x, v);
					}
				}

			}

		}
		relations = ret;
	}

	private double closure(int a, int x, int y) {
		double val = relations[a].getCell(x, y);// 0

		for (int o = 0; o < n_objects; o++) {
			val = Math.max(FuzzySet.compose(relations[a].getCell(x, o),
					relations[a].getCell(o, y)), val);
			if (val == 1)
				break;
		}
		return val;
	}

	public final void setAttributeStats() {
		attrDifference = new double[n_attributes]; // difference between max
		// and min
		attrSTD = new double[n_attributes]; // standard deviation
		mean = new double[n_attributes]; // mean
		max = new double[n_attributes];
		min = new double[n_attributes];
		double sum_sqr = 0;
		double std = 0;
		double n, sum;
		double value;

		for (int a = 0; a < n_attributes; a++) {
			max[a] = -99999999999.0;
			min[a] = 99999999999.0;
			sum_sqr = 0;
			sum = 0;
			n = 0;

			for (int o = 0; o < n_objects; o++) {
				value = fm.getCell(o, a);
				if (Double.isNaN(value)) {}
				else {
					if (value > max[a])
						max[a] = value;
					if (value < min[a])
						min[a] = value;
					sum_sqr += value * value;
					sum += value;
					n++;
				}
			}

			std = Math.sqrt(((n * sum_sqr) - (sum * sum)) / (n * (n - 1)));

			mean[a] = sum / n;
			attrDifference[a] = Math.abs(max[a] - min[a]);
			if (attrDifference[a] == 0)
				attrDifference[a] = 0.0000001;
			attrSTD[a] = std;
			if (std == 0)
				attrSTD[a] = 0.00001;

			// System.err.println("stdev for "+a+": "+std+" "+max[a]+",
			// "+min[a]);
		}
	}

	// Fuzzy similarity for type-1 FRFS (m == 1)
	private final double fuzzySimilarity(int attr, double mainVal,
			double otherVal) {
		return fuzzySimilarity(attr, mainVal, otherVal, 1);
	}

	// Fuzzy similarity m is used for IV-FRFS
	private final double fuzzySimilarity(int attr, double mainVal,
			double otherVal, double m) {
		double ret = 0;
		int sim = -1;
		double al = 0;
		double be = 0;

		if (attr == n_attributes - 1) {
			sim = frfs.dec_similarity; // if this is the decision feature, use
			// the decision similarity measure
			al = frfs.alpha_d;
			be = frfs.beta_d;
		} else {
			sim = SIMILARITY;
			al = frfs.alpha_c;
			be = frfs.beta_c;
		}

		switch (sim) {
		case GAUSSIAN:
			ret = Math.exp(-0.5 * ((mainVal - otherVal) * (mainVal - otherVal))
					/ ((1 * attrSTD[attr] * attrSTD[attr])));
			break;
		case TRIANGULAR:
			ret = Math.max(0, 1 - Math.pow(
					(Math.abs(mainVal - otherVal) / attrDifference[attr]), m));
			break;
		case TRIANGULAR2:
			ret = Math.max(Math.min((otherVal - mainVal + m * attrSTD[attr])
					/ (m * attrSTD[attr]),
					(mainVal + m * attrSTD[attr] - otherVal)
					/ (m * attrSTD[attr])), 0);
			break;
		case EQUIVALENCE: // crisp equivalence
			if (mainVal == otherVal)
				ret = 1;
			else
				ret = 0;
			break;

		case TRIANGULAR3:// cornelis et al
			ret = Math.max(1 - Math.abs(mainVal - otherVal), 0);
			break;
		case CORNELISA:
			ret = Math.max(Math.min(1,be - al* (Math.abs(mainVal - otherVal) / attrDifference[attr])),0);
			break;
		}
		return ret;
	}

	// lower approximation for x
	private final double lowerApprox(Relation rel, int x, int d) {
		double val = 0;
		double val2 = 1;

		// for each fuzzy equivalence class in the relation
		for (int f = 0; f < n_objects; f++) {
			val2 = 1;
			for (int y = 0; y < n_objects; y++) {
				if (frfs.OPTIMISE >= 2)
					val2 = Math.min(FuzzySet.impl(rel.getCell(f, y), 1),
							fuzzySimilarity(n_attributes - 1, fm.getCell(d,
									n_attributes - 1), fm.getCell(y,
											n_attributes - 1)));
				else
					val2 = Math.min(FuzzySet.impl(rel.getCell(f, y), 1),
							relations[n_attributes - 1].getCell(d, y));

				if (val2 == 0)
					break;
			}

			val = Math.max(FuzzySet.tnorm(rel.getCell(f, x), val2), val);
			if (val == 1)
				break;
		}
		return val;
	}

	// original FRFS
	private final double calculateGamma4(Relation rel, int a) {
		double ret = 0;

		for (int x = 0; x < n_objects; x++) {
			double val = 0;

			// positive region
			for (int d = 0; d < n_objects; d++) {
				val = Math.max(lowerApprox(rel, x, d), val);
				if (val == 1)
					break;
			}
			ret += val;
			if ((bestGamma * n_objects) > ret + (n_objects - x))
				break;
		}
		return ret / n_objects_d;
	}

	private final void setIndexes(Relation rel) {
		for (int i = 0; i < n_objects; i++) {
			boolean same = false;

			for (int j = i + 1; j < n_objects; j++) {

				for (int d = 0; d < n_objects; d++) {
					if (rel.getCell(i, d) != rel.getCell(j, d)) {
						same = false;
						break;
					} else
						same = true;
				}
				if (same) {
					indexes[j] = i + 1;
					break;
				} // i+1 as default is 0
			}

		}
	}

	// VQRS2 - weak version
	private final double calculateVQRS2(Relation rel) {
		double ret = 0;

		for (int x = 0; x < n_objects; x++) {
			double val = 0;

			val = lowerVQRS(rel, x, x); // use object x's decision class only
			ret += val;

		}
		return Math.min(1, ret / n_objects_d);
	}

	// VQRS
	private final double calculateVQRS(Relation rel) {
		double ret = 0;
		// for optimization
		indexes = new int[n_objects];
		vals = new double[n_objects];

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		// if (frfs2.dec_similarity==EQUIVALENCE)
		setIndexes(rel);

		for (int x = 0; x < n_objects; x++) {
			double val = 0;

			if (indexes[x] != 0) {
				vals[x] = vals[indexes[x] - 1];
				ret += vals[x];
			} else {
				decVals = new double[n_objects];

				// positive region
				for (int d = 0; d < n_objects; d++) {
					if (decIndexes[d] != 0) {
						decVals[d] = decVals[decIndexes[d] - 1];
					} else {
						decVals[d] = lowerVQRS(rel, x, d);
						val = Math.max(decVals[d], val);
					}
					if (val == 1)
						break;
				}
				vals[x] = val;
				ret += val;
			}
		}
		return Math.min(1, ret / n_objects_d);
	}

	private final double getDecisionSimilarity(Relation rel, int x, int f) {
		double ret=0;

		switch(frfs.OPTIMISE) {
		case 0: case 1:
			ret=relations[n_attributes - 1].getCell(x, f);
			break;
		case 2: case 3: case 4:
			ret=fuzzySimilarity(
					n_attributes - 1, fm.getCell(x, n_attributes - 1), fm
					.getCell(f, n_attributes - 1));
			break;
		}
		return ret;
	}
	private final double getSimilarity(Relation rel, int x, int f, Subset reduct) {
		double ret=0;

		//for conditional attributes

		switch(frfs.OPTIMISE) {
		case 0: case 1: case 2: case 3:
			ret=rel.getCell(x, f);
			break;
		case 4:
			ret=1;
			for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
				ret = FuzzySet.compose(fuzzySimilarity(a, fm.getCell(x, a), fm
						.getCell(f, a)), ret);
				if (ret==0) break;

			}
			break;
		}

		return ret;
	}

	private final double lowerVQRS(Relation rel, int x, int d) {
		double val = 0;
		double denom = 0;

		// for each fuzzy equivalence class in the relation
		for (int f = 0; f < n_objects; f++) {
			double condSim = getSimilarity(rel, x, f, Reduct);
			val += FuzzySet.tnorm(condSim, getDecisionSimilarity(rel, d, f));

			denom += condSim;

		}
		return Q(val / denom, alpha_u, beta_u); // return quantified value
	}

	// fuzzy quantifier
	private final double Q(double x, double alpha, double beta) {
		double ret = 0;
		denomVal = (beta - alpha) * (beta - alpha);

		if (x <= alpha)
			ret = 0;
		else if (x < ((alpha + beta) / 2)) {
			ret = (2 * (x - alpha) * (x - alpha)) / denomVal;
		} else if (x < beta) {
			ret = 1 - ((2 * (x - beta) * (x - beta)) / denomVal);
		} else if (beta <= x) {
			ret = 1;
		}

		return ret;

	}

	// produce a random reduct of size 'numFeatures'
	public final Subset randomReduct(int numFeatures) {
		Subset ret = new Subset(n_attributes);
		ret.add(n_attributes - 1);
		Random random = new Random();
		int feature;

		for (int i = 0; i < numFeatures; i++) {
			feature = Math.abs(random.nextInt()) % (n_attributes - 1);
			if (ret.contains(feature))
				i--; // already got this so try again
			else
				ret.add(feature);
		}
		System.err.println("Random reduct is " + ret + " |" + ret.cardinality()
				+ "|");
		return ret;
	}

	// gain ratio (fuzzy: fuztree.pdf)
	private final double calculateGR(Relation rel, int a) {
		double ret = 0;
		double cardinality = 0;
		double IV = 0;

		for (int i = 0; i < n_objects; i++) {

			for (int j = 0; j < n_objects; j++) {
				cardinality += rel.getCell(i, j);
			}

		}

		// calculate Yi
		for (int x = 0; x < n_objects; x++) {
			cards[x] = 0;
			for (int o = 0; o < n_objects; o++) {
				cards[x] += getSimilarity(rel, x, o, Reduct);
			}
			double temp = cards[x] / cardinality;// n_objects
			IV += -temp * log2(temp);
		}

		cardinality = 0;

		// for optimization
		indexes = new int[n_objects];
		vals = new double[n_objects];

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);
		double temp;

		for (int x = 0; x < n_objects; x++) {
			temp = 0;
			// System.out.print(indexes[x]+" ");
			if (indexes[x] != 0) {
				cards[x] = cards[indexes[x] - 1];
				vals[x] = vals[indexes[x] - 1];
			} else {
				decVals = new double[n_objects];
				for (int d = 0; d < n_objects; d++) {
					if (decIndexes[d] != 0) {
						decVals[d] = decVals[decIndexes[d] - 1];
						temp += decVals[d];
					} else {
						double val = 0;
						double denom = 0;// temp=0;

						for (int f = 0; f < n_objects; f++) {
							double condSim = getSimilarity(rel, x, f, Reduct);

							val += FuzzySet.tnorm(
									condSim,
									getDecisionSimilarity(rel, d, f));


							denom += condSim;

						}
						cards[x] = denom;

						decVals[d] = (-(val / denom) * log2(val / denom))
								/ decisionLength[d];
						temp += decVals[d];

						// optimization
						if (cards[x] == 1 && frfs.dec_similarity == EQUIVALENCE)
							break;
					}
				}

				vals[x] = cards[x] * temp;

			}
			cardinality += cards[x];
			ret += vals[x];

		}
		ret /= cardinality;
		entropy = ret / log2(decCardinality);

		return 1 - (ret / IV);// (prevEntropy - entropy)/IV;
	}

	// Entropy
	private final double calculateEntropy(Relation rel, int a) {
		double ret = 0;
		double temp = 0;
		double cardinality = 0;

		// for optimization
		indexes = new int[n_objects];
		vals = new double[n_objects];

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);

		for (int x = 0; x < n_objects; x++) {
			temp = 0;
			// System.out.print(indexes[x]+" ");
			if (indexes[x] != 0) {
				cards[x] = cards[indexes[x] - 1];
				vals[x] = vals[indexes[x] - 1];
			} else {
				decVals = new double[n_objects];
				for (int d = 0; d < n_objects; d++) {
					if (decIndexes[d] != 0) {
						decVals[d] = decVals[decIndexes[d] - 1];
						temp += decVals[d];
					} else {
						double val = 0;
						double denom = 0;// temp=0;

						for (int f = 0; f < n_objects; f++) {
							double condSim = getSimilarity(rel, x, f, Reduct);
							val += FuzzySet.tnorm(
									condSim,
									getDecisionSimilarity(rel, d, f));


							denom += condSim;

						}
						cards[x] = denom;

						decVals[d] = (-(val / denom) * log2(val / denom))
								/ decisionLength[d];
						temp += decVals[d];

						// optimization (for crisp decisions)
						if (cards[x] == 1 && frfs.dec_similarity == EQUIVALENCE)
							break;

					}
				}

				vals[x] = cards[x] * temp;

			}
			cardinality += cards[x];
			ret += vals[x];
		}
		ret /= cardinality;
		return 1 - (ret / log2(decCardinality));
	}

	public static final double log2(double x) {
		if (x > 0)
			return Math.log(x) / LN2;
		else
			return 0.0;
	}

	private final double calculateGamma(Relation rel, int a) {
		double lower;
		double currLower = 0;
		double ret = 0;
		double val = 0;
		double sim = 1;

		// for optimization
		indexes = new int[n_objects];
		vals = new double[n_objects];

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);

		for (int x = 0; x < n_objects; x++) {
			if (indexes[x] != 0) {
				vals[x] = vals[indexes[x] - 1];
			} else {
				val = 0;
				decVals = new double[n_objects];
				for (int d = 0; d < n_objects; d++) {
					if (decIndexes[d] != 0) {
						decVals[d] = decVals[decIndexes[d] - 1];
					} else {
						lower = 1;

						if (frfs.dec_similarity == EQUIVALENCE) 
							sim = getDecisionSimilarity(rel,d,x);



						// only non-zero if real-valued decisions or identical
						// crisp decisions
						if (sim != 0) {
							// lower approximations of object x
							for (int y = 0; y < n_objects; y++) {
								double condSim = getSimilarity(rel, x, y, Reduct);

								currLower = FuzzySet.impl(
										condSim, 
										getDecisionSimilarity(rel, d, y));

								//if (d==3&&x==1) System.err.println(y+" => "+currLower);

								lower = Math.min(currLower, lower);
								if (lower == 0)
									break;
							}
							decVals[d] = lower;
							val = Math.max(lower, val);
						}
					}
					// if (d==3&&x==1) System.err.println(d+" => "+decVals[d]);

					if (val == 1)
						break;
					if (frfs.dec_similarity == EQUIVALENCE && sim == 1)
						break; // Optimization: can't better this value in
					// crisp domain
				}

				vals[x] = val;

			}

			ret += vals[x];
		}

		return ret / n_objects_d;
	}

	// weak version
	private final double calculateGamma2(Relation rel, int a) {
		double lower;
		double currLower = 0;
		double ret = 0;
		int d = -1;

		// if (a==3||a==4) System.err.println("\n-----------------\n"+a);

		for (int x = 0; x < n_objects; x++) {
			d = x;// weak version, so only consider x's decision
			lower = 1;
			// if (a==2) System.err.println("For object: "+x);

			// lower approximations of object x
			for (int y = 0; y < n_objects; y++) {
				// if (a==3||a==4) System.err.println("("+x+","+y+") =
				// "+rel.getCell(x,y));

				double condSim = getSimilarity(rel, x, y, Reduct);

				currLower = FuzzySet.impl(
						condSim, 
						getDecisionSimilarity(rel, d, y));

				// System.err.println("("+x+","+y+") = "+currLower);
				lower = Math.min(currLower, lower);

				// if (a==2) System.err.println(y+ " => "+ currLower);

				if (lower == 0)
					break;
			}
			// if (a==3||a==4) System.err.println(x+ " => "+ lower);

			// if (a==2) System.err.println(" => "+ lower);

			ret += lower;

		} // if (a==3||a==4) System.err.println("\n-----------------\n");
		return ret / n_objects_d;
	}

	private final double calculateBoundary(Relation rel, int a) {
		double lower;
		lowers = new double[n_objects];
		double currLower = 0;
		uppers = new double[n_objects];
		double ret = 0;
		double upper = 0;
		double currUpper = 0;
		double denom = 0;
		denom = (n_objects_d * decCardinality);

		// for optimization
		indexes = new int[n_objects];
		vals = new double[n_objects];

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);

		for (int x = 0; x < n_objects; x++) {
			if (indexes[x] != 0) {
				vals[x] = vals[indexes[x] - 1];
				ret += vals[x];
			} else {
				for (int d = 0; d < n_objects; d++) {
					if (decIndexes[d] != 0) {
						lowers[d] = lowers[decIndexes[d] - 1];
					} else {
						lower = 1;

						// lower approximations of object x
						for (int y = 0; y < n_objects; y++) {
							double condSim = getSimilarity(rel, x, y, Reduct);

							currLower = FuzzySet.impl(
									condSim, 
									getDecisionSimilarity(rel, d, y));

							lower = Math.min(currLower, lower);
							if (lower == 0)
								break;
						}

						lowers[d] = lower;
					}
				}

				double temp = 0;

				for (int d = 0; d < n_objects; d++) {
					if (lowers[d] == 1) {
						uppers[d] = 1;
					} else {
						if (decIndexes[d] != 0) {
							uppers[d] = uppers[decIndexes[d] - 1];
						} else {
							upper = 0;

							// upper approximations of object x
							for (int y = 0; y < n_objects; y++) {
								double condSim = getSimilarity(rel, x, y, Reduct);

								currUpper = FuzzySet.tnorm(
										condSim, 
										getDecisionSimilarity(rel, d, y));

								upper = Math.max(currUpper, upper);
								if (upper == 1)
									break;
							}

							uppers[d] = upper;
						}

						temp += (uppers[d] - lowers[d]) / decisionLength[d];
					}
				}
				vals[x] = temp;
				ret += temp;
			}

			// optimization - if we're not going to do better than bestGamma,
			// then quit
			// if ((ret/val)+(bestGamma*c_divisor)>1) {ret=val;break;}
		}

		return 1 - (ret / denom);
	}

	private final double calcEntropy(double sumBoundary, double[] boundaryMems) {
		double value = 0;

		for (int x = 0; x < n_objects; x++) {
			if (boundaryMems[x] == 0) {
			} else {
				value += -(boundaryMems[x] / sumBoundary)
						* log2(boundaryMems[x] / sumBoundary);
			}
		}
		return value;
	}

	private final double calculateBoundary2(Relation rel, int a) {
		double lower;
		double upper = 0;
		double currLower = 0;
		double currUpper = 0;
		double ret = 0;
		double denom = 0;

		denom = (n_objects_d * decCardinality);

		// for optimization
		decVals = new double[n_objects];
		indexes = new int[n_objects];
		double temp = 0;

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);

		for (int d = 0; d < n_objects; d++) {
			if (decIndexes[d] != 0) {
				decVals[d] = decVals[decIndexes[d] - 1];
				ret += decVals[d];
			} else {
				temp = 0;
				vals = new double[n_objects];
				for (int x = 0; x < n_objects; x++) {
					if (indexes[x] != 0) {
						vals[x] = vals[indexes[x] - 1];
						temp += vals[x];
					} else {
						lower = 1;
						upper = 0;

						// lower and upper approximations of decision d for
						// object x
						for (int y = 0; y < n_objects; y++) {
							double condSim = getSimilarity(rel, x, y, Reduct);
							double decSim = getDecisionSimilarity(rel, d, y);

							currLower = FuzzySet.impl(condSim, decSim);
							currUpper = FuzzySet.tnorm(condSim, decSim);

							lower = Math.min(currLower, lower);
							upper = Math.max(currUpper, upper);
							if (upper == 1 && lower == 0)
								break;

						}
						vals[x] = (upper - lower) / decisionLength[d];
						temp += vals[x];
					}
				}
				decVals[d] = temp;
				ret += decVals[d];

			}
			// opt = (ret/val)+(bestGamma*c_divisor);
			// if (!verbose&&opt>1) {ret=val;break;}
		}

		return 1 - (ret / denom);
	}

	private final double calculateBoundaryEnt(Relation rel, int a) {
		double lower;
		double upper = 0;
		double currLower = 0;
		double currUpper = 0;
		double ret = 0;

		double[] boundaryMems;

		decVals = new double[n_objects];
		// for optimization
		indexes = new int[n_objects];
		double temp = 0;

		double normalization = log2(n_objects);

		// optimization - see if other objects in the dataset are identical/have
		// identical similarity
		setIndexes(rel);

		for (int d = 0; d < n_objects; d++) {
			boundaryMems = new double[n_objects];

			if (decIndexes[d] != 0) {
				decVals[d] = decVals[decIndexes[d] - 1];
				ret += decVals[d];
			} else {
				temp = 0;
				vals = new double[n_objects];
				for (int x = 0; x < n_objects; x++) {
					if (indexes[x] != 0) {
						vals[x] = vals[indexes[x] - 1];
						boundaryMems[x] = vals[x];
						temp += vals[x];
					} else {
						lower = 1;
						upper = 0;

						// lower and upper approximations of decision d for
						// object x
						for (int y = 0; y < n_objects; y++) {
							double condSim = getSimilarity(rel, x, y, Reduct);
							double decSim = getDecisionSimilarity(rel, d, y);

							currLower = FuzzySet.impl(condSim, decSim);
							currUpper = FuzzySet.tnorm(condSim, decSim);

							lower = Math.min(currLower, lower);
							upper = Math.max(currUpper, upper);
							if (upper == 1 && lower == 0)
								break;

						}
						vals[x] = (upper - lower) / decisionLength[d];
						boundaryMems[x] = vals[x];// *decisionLength[d];
						temp += vals[x];
						// if (1 - (ret/val)<bestGamma) break;
					}
				}
				decVals[d] = calcEntropy(temp, boundaryMems)
						/ (normalization * decisionLength[d]);
				ret += decVals[d];
			}

		}

		// return 1 - (ret/(double)(n_objects*decCardinality));
		return 1 - (ret / decCardinality);
	}

	private final double roughEntropy(Relation rel, int a) {
		double lower;
		double currLower = 0;
		double ret = 0;
		double val = 0;
		double[] posRegion = new double[n_objects];
		double val2 = 0;

		for (int d = 0; d < n_objects; d++) {

			for (int x = 0; x < n_objects; x++) {
				lower = 1;
				val = 0;

				// lower approximations of object x
				for (int y = 0; y < n_objects; y++) {
					double condSim = getSimilarity(rel, x, y, Reduct);
					double decSim = getDecisionSimilarity(rel, d, y);

					currLower = FuzzySet.impl(condSim, decSim);

					lower = Math.min(currLower, lower);
					val += rel.getCell(x, y);
				}
				val2 += (val / n_objects_d) * log2(val / n_objects_d);
				posRegion[x] = Math.max(lower, posRegion[x]);
			}
		}
		for (int o = 0; o < n_objects; o++)
			ret += posRegion[o];
		return ret;
	}

	// slezak's measure based on minimum positive region membership
	private final double calculateDelta(Relation rel, int a) {
		double lower;
		double currLower = 0;
		double ret = 1;
		double val = 0;

		for (int x = 0; x < n_objects; x++) {
			val = 0;

			if (frfs.dec_similarity != EQUIVALENCE) {
				for (int d = 0; d < n_objects; d++) {
					lower = 1;

					for (int y = 0; y < n_objects; y++) {
						double condSim = getSimilarity(rel, x, y, Reduct);
						double decSim = getDecisionSimilarity(rel, d, y);

						currLower = FuzzySet.impl(condSim, decSim);

						lower = Math.min(currLower, lower);
						if (lower == 0)
							break;
					}

					val = Math.max(lower, val);
					if (val == 1)
						break;

				}
			} else { // optimized version for crisp decisions
				int d = x;
				lower = 1;

				for (int y = 0; y < n_objects; y++) {
					double condSim = getSimilarity(rel, x, y, Reduct);
					double decSim = getDecisionSimilarity(rel, d, y);

					currLower = FuzzySet.impl(condSim, decSim);

					lower = Math.min(currLower, lower);
					if (lower == 0)
						break;
				}

				val = lower;

			}
			// System.err.println(x+": "+val);
			ret = Math.min(ret, val);
			if (ret == 0)
				break;
		}

		return ret;
	}

	// slezak's measure based on minimum positive region membership  - weak version
	private final double calculateDelta2(Relation rel, int a) {
		double lower;
		double currLower = 0;
		double ret = 1;
		double val = 0;

		for (int x = 0; x < n_objects; x++) {
			val = 0;

			int d = x;
			lower = 1;

			for (int y = 0; y < n_objects; y++) {
				double condSim = getSimilarity(rel, x, y, Reduct);
				double decSim = getDecisionSimilarity(rel, d, y);

				currLower = FuzzySet.impl(condSim, decSim);

				lower = Math.min(currLower, lower);
				if (lower == 0) break;
			}

			val = lower;

			// System.err.println(x+": "+val);
			ret = Math.min(ret, val);
			if (ret == 0) break;
		}

		return ret;
	}


	private final void generatePartition(Relation curr, int a, Subset reduct) {

		switch (frfs.OPTIMISE) {
		case 0:
			generatePartition(reduct);
			break;
		case 1:
			generatePartition(curr, a);
			break;
		case 3:
			generatePartition3(reduct);
			break;
		case 4: break;
		default:
			generatePartition2(curr, a);
			break;
		}

	}

	int nextBit = 0;

	private final void generatePartition(Subset reduct) {
		nextBit = reduct.nextFeature(0);

		for (int i = 0; i < n_objects; i++) {
			temprel.setCell(i, i, 1);

			for (int j = i + 1; j < n_objects; j++) {
				rel = relations[nextBit].getCell(i, j);
				temprel.setCell(i, j, rel);
				temprel.setCell(j, i, rel);
			}
		}

		if (reduct.cardinality() == 1) {
		} else {
			for (int o = 0; o < n_objects; o++) {
				for (int o1 = o + 1; o1 < n_objects; o1++) {
					for (int a = reduct.nextFeature(nextBit + 1); a >= 0; a = reduct
							.nextFeature(a + 1)) {
						rel = FuzzySet.compose(relations[a].getCell(o, o1),
								temprel.getCell(o, o1));
						temprel.setCell(o, o1, rel);
						temprel.setCell(o1, o, rel);
						if (rel == 0)
							break;
					}
				}
			}
		}

	}

	double rel = 1;

	// further optimized version - don't store relations for each
	// attribute, work them out on the fly - however, this does mean
	// that transitive closures can't be calculated and hence a
	// suitable similarity measure should be used
	private final void generatePartition(Relation curr, int a) {

		for (int o = 0; o < n_objects; o++) {
			temprel.setCell(o, o, 1);

			for (int o1 = o + 1; o1 < n_objects; o1++) {
				rel = FuzzySet.compose(relations[a].getCell(o, o1), curr
						.getCell(o, o1));
				temprel.setCell(o, o1, rel);
				temprel.setCell(o1, o, rel);
			}
		}
	}

	// optimised version: opt 2
	private final void generatePartition2(Relation curr, int a) {
		double mainValue;

		for (int o = 0; o < n_objects; o++) {
			temprel.setCell(o, o, 1);
			mainValue = fm.getCell(o, a);

			for (int o1 = o + 1; o1 < n_objects; o1++) {
				rel = FuzzySet.compose(fuzzySimilarity(a, mainValue, fm
						.getCell(o1, a)), curr.getCell(o, o1));
				temprel.setCell(o, o1, rel);
				temprel.setCell(o1, o, rel);
			}
		}
	}

	private final void generatePartition3(Subset reduct) {
		nextBit = reduct.nextFeature(0);
		double mainValue = 0;

		for (int i = 0; i < n_objects; i++) {
			temprel.setCell(i, i, 1);
			mainValue = fm.getCell(i, nextBit);

			for (int j = i + 1; j < n_objects; j++) {
				rel = fuzzySimilarity(nextBit, mainValue, fm
						.getCell(j, nextBit));
				temprel.setCell(i, j, rel);
				temprel.setCell(j, i, rel);
			}
		}

		if (reduct.cardinality() == 1) {
		} else {
			for (int o = 0; o < n_objects; o++) {
				for (int o1 = o + 1; o1 < n_objects; o1++) {
					for (int a = reduct.nextFeature(nextBit + 1); a >= 0; a = reduct
							.nextFeature(a + 1)) {
						rel = FuzzySet.compose(fuzzySimilarity(a, fm.getCell(o,
								a), fm.getCell(o1, a)), temprel.getCell(o, o1));
						temprel.setCell(o, o1, rel);
						temprel.setCell(o1, o, rel);
						if (rel == 0)
							break;
					}
				}
			}
		}
	}

	private final void initCurrent() {
		if (frfs.OPTIMISE!=4) {
			for (int i = 0; i < n_objects; i++) {
				for (int j = i; j < n_objects; j++) {
					current.setCell(i, j, 1);
					current.setCell(j, i, 1);
				}
			}
		}
	}

	private final double[] simT2(int x, int y, Subset reduct) {
		double[] ret = new double[2];
		ret[0] = 1;
		ret[1] = 1;

		if (x==y) return ret;

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			double val1 = fm.getCell(x, a);
			double val2 = fm.getCell(y, a);

			//if (val1 == frfs.MISSING || val2 == frfs.MISSING) {
			if (Double.isNaN(val1) || Double.isNaN(val2)) {
				ret[0] = 0;
				// ret[1] = 1
				// System.err.println("blah");
				// break;
			} else {
				ret[0] = FuzzySet.compose(
						fuzzySimilarity(a, val1, val2, frfs.m), ret[0]);
				ret[1] = FuzzySet.compose(fuzzySimilarity(a, val1, val2, 1),
						ret[1]);
			}
		}
		// System.err.println(ret[0]+" "+ret[1]);
		return ret;
	}

	public double fdmT2(Subset reduct) {
		double[] difd = new double[2];
		double[] val = new double[2];

		double[] ret = new double[2];
		ret[0] = 0;
		ret[1] = 0;

		double[] valueRA = new double[2];


		for (int i = 0; i < n_objects - 1; i++) {
			for (int j = i + 1; j < n_objects; j++) {

				// decision attribute values
				difd[0] = fuzzySimilarity(n_attributes - 1, fm.getCell(
						i, n_attributes - 1), fm.getCell(j,
								n_attributes - 1), frfs.m);
				difd[1] = fuzzySimilarity(n_attributes - 1, fm.getCell(
						i, n_attributes - 1), fm.getCell(j,
								n_attributes - 1), 1);

				valueRA = simT2(i, j, reduct);

				val[0] = FuzzySet.impl(valueRA[1],difd[0]);
				val[1] = FuzzySet.impl(valueRA[0],difd[1]);

				ret[0] += val[0];
				ret[1] += val[1];
			}
		}
		//return ret / ((n_objects_d * n_objects_d - n_objects_d) / 2);
		return (ret[0]+ret[1])/(n_objects_d * n_objects_d - n_objects_d);
	}

	// interval-valued FRFS
	private final double calculateGammaT2(Subset reduct) {
		double lower1;
		double lower2;
		double currLower1 = 0;
		double currLower2 = 0;
		double[] val = new double[2];
		val[0] = 0;
		val[1] = 0; // positive region
		double[] dep = new double[2];
		dep[0] = 0;
		dep[1] = 0; // dependency

		double[] sim = new double[2];
		double[][] decVals;
		// for optimization
		indexes = new int[n_objects];
		double[][] vals = new double[n_objects][2];
		double[] tempSim;

		// setIndexes
		/*
		 * for (int i=0;i<n_objects;i++) { boolean same=false;
		 * 
		 * for (int j=i+1;j<n_objects;j++) {
		 * 
		 * for (int a=reduct.nextSetBit(0);a>=0;a=reduct.nextSetBit(a+1)) {
		 * 
		 * if (fm.getCell(i,a)!=fm.getCell(j,a)) { same=false; break; } else
		 * same=true; } if (same) {indexes[j]=i+1;break;} //i+1 as default is 0 }
		 *  }
		 */

		for (int x = 0; x < n_objects; x++) {
			// if (indexes[x]!=0)
			// {vals[x][0]=vals[indexes[x]-1][0];vals[x][1]=vals[indexes[x]-1][1];}
			// else {
			val[0] = 0;
			val[1] = 0;
			decVals = new double[n_objects][2];
			for (int d = 0; d < n_objects; d++) {
				if (decIndexes[d] != 0) {
					decVals[d][0] = decVals[decIndexes[d] - 1][0];
					decVals[d][1] = decVals[decIndexes[d] - 1][1];
				} else {
					lower1 = 1;
					lower2 = 1;

					// lower approximations of object x
					for (int y = 0; y < n_objects; y++) {
						tempSim = simT2(x, y, reduct);

						sim[0] = fuzzySimilarity(n_attributes - 1, fm.getCell(
								d, n_attributes - 1), fm.getCell(y,
										n_attributes - 1), frfs.m);
						sim[1] = fuzzySimilarity(n_attributes - 1, fm.getCell(
								d, n_attributes - 1), fm.getCell(y,
										n_attributes - 1), 1);

						currLower1 = FuzzySet.impl(tempSim[1], sim[0]);
						currLower2 = FuzzySet.impl(tempSim[0], sim[1]);

						lower1 = Math.min(currLower1, lower1);
						lower2 = Math.min(currLower2, lower2);
						if (lower1 == 0 && lower2 == 0) break;
					}

					decVals[d][0] = lower1;
					decVals[d][1] = lower2;
					val[0] = Math.max(lower1, val[0]);
					val[1] = Math.max(lower2, val[1]);
					if (val[0] == 1 && val[1] == 1) break;

				}

			}
			vals[x][0] = val[0];
			vals[x][1] = val[1];
			// }
			dep[0] += vals[x][0];
			dep[1] += vals[x][1];
		}
		// System.err.println(dep[0]+" "+dep[1]);
		dep[0] /= (n_objects_d);
		dep[1] /= (n_objects_d);

		return (dep[0] + dep[1]) / 2;
	}

	// interval-valued FRFS
	private final double calculateBoundaryT2(Subset reduct) {
		double lower1,upper1;
		double lower2,upper2;

		double[] val = new double[2];
		val[0] = 0;
		val[1] = 0; // positive region
		double[] dep = new double[2];
		dep[0] = 0;
		dep[1] = 0; // dependency

		double[] sim = new double[2];
		double[][] decVals;
		// for optimization
		indexes = new int[n_objects];

		double[] tempSim;
		double denom = (n_objects_d * decCardinality);


		for (int x = 0; x < n_objects; x++) {

			val[0] = 0;
			val[1] = 0;
			decVals = new double[n_objects][2];

			for (int d = 0; d < n_objects; d++) {
				if (decIndexes[d] != 0) {
					decVals[d][0] = decVals[decIndexes[d] - 1][0];
					decVals[d][1] = decVals[decIndexes[d] - 1][1];
				} else {
					lower1 = 1;
					lower2 = 1;
					upper1 = 0;
					upper2 = 0;

					// lower approximations of object x
					for (int y = 0; y < n_objects; y++) {
						tempSim = simT2(x, y, reduct);

						sim[0] = fuzzySimilarity(n_attributes - 1, fm.getCell(
								d, n_attributes - 1), fm.getCell(y,
										n_attributes - 1), frfs.m);
						sim[1] = fuzzySimilarity(n_attributes - 1, fm.getCell(
								d, n_attributes - 1), fm.getCell(y,
										n_attributes - 1), 1);

						lower1 = Math.min(FuzzySet.impl(tempSim[1], sim[0]), lower1);
						lower2 = Math.min(FuzzySet.impl(tempSim[0], sim[1]), lower2);

						upper1 = Math.max(FuzzySet.tnorm(tempSim[0], sim[0]),upper1);
						upper2 = Math.max(FuzzySet.tnorm(tempSim[1], sim[1]),upper2);

						//lower2 is the upper bound for min
						//upper1 is the lower bound for max
						if (lower2==0 && upper1 == 1) break;
					}

					decVals[d][0] = (upper1-lower1) /decisionLength[d];
					decVals[d][1] = (upper2-lower2) /decisionLength[d];


				}
				val[0]+=decVals[d][0];
				val[1]+=decVals[d][1];
			}

			dep[0] += val[0];
			dep[1] += val[1];
		}

		//return 1 - (dep[0] + dep[1])/2;
		return 1-((dep[0]/denom + dep[1]/denom) / 2);
	}

	public final void qsba(Matrix data2, int dObjects, Subset reduct) {
		// output a rule for each decision concept
		for (int d = 0; d < decCardinality; d++) {

		}

	}

	// classify objects in data2 (contains dObjects number of objects)
	public final void classify(Matrix data2, int dObjects, Subset reduct) {
		double accuracy = 0;

		Subset full = new Subset(n_attributes);
		for (int i = 0; i < n_attributes - 1; i++)
			full.add(i);

		if (reduct == null)
			reduct = full;
		if (reduct.contains(n_attributes - 1))
			reduct.remove(n_attributes - 1);

		// generate partition for full feature set

		BitSet remaining = new BitSet(n_objects); // a set containing all
		// objects
		for (int i = 0; i < n_objects; i++)
			remaining.set(i);

		BitSet eq;
		// create the ind classes for the decision attributes
		for (int o1 = 0; o1 < n_objects; o1++) {
			if (remaining.get(o1)) {
				eq = new BitSet(n_objects);
				// add this object (o1) to the current set
				eq.set(o1);
				for (int o2 = o1 + 1; o2 < n_objects; o2++)
					if (remaining.get(o2))
						if (fm.getCell(o1, n_attributes - 1) == fm.getCell(o2,
								n_attributes - 1))
							eq.set(o2);
				for (int i = 0; i < n_objects; i++)
					if (eq.get(i))
						remaining.clear(i);
				decision.add((BitSet) eq.clone());
			}
		}

		int classification;
		ArrayList<Index> objectIndex;
		BitSet dec;
		int[][] confusion = new int[decision.size()][decision.size()];
		//map from decision index to actual class value
		double[] classValues = new double[decision.size()];

		//initialise confusion matrix
		if (frfs.outputConfusionMatrix) {
			for (int i=0;i<decision.size();i++) {
				for (int j=i;j<decision.size();j++) {
					confusion[i][j] = 0;
					confusion[j][i] = 0;
				}
				classValues[i] = fm.getCell((decision.get(i)).nextSetBit(0), n_attributes - 1); 
			}

		}

		for (int i = 0; i < dObjects; i++) {
			objectIndex = new ArrayList<Index>(n_objects);
			classification = -2;

			// make the classification
			switch (frfs.command) {
			case frfs.NN:
			case frfs.VQNN: // fuzzy-rough NN
				// compute similarities (for nearest neighbours)
				for (int o = 0; o < n_objects; o++) {
					objectIndex.add(new Index(o, sim(o, i, data2, reduct)));
				}

				// sort objects according to their similarity to object i
				Collections.sort(objectIndex);
				classification = classify2(i, data2, reduct, objectIndex);

				break;
			case frfs.FNN:
				// compute similarities (for nearest neighbours)
				for (int o = 0; o < n_objects; o++) {
					objectIndex.add(new Index(o, sim2(o, i, data2, reduct)));
				}

				// sort objects according to their similarity to object i
				Collections.sort(objectIndex);
				classification = classify2fnn(i, data2, reduct, objectIndex);
				break;
			case frfs.FRNN:// fuzzy-rough ownership knn
				classification = classify2Ownership(i, data2, reduct, objectIndex);
			}

			dec = decision.get(classification);
			if (!frfs.visualise && fm.getCell(dec.nextSetBit(0), n_attributes - 1) == data2.getCell(i, n_attributes - 1))
				accuracy++;

			if (frfs.outputConfusionMatrix) {
				int actual=-1;
				for (actual=0;actual<decision.size();actual++) {
					if (data2.getCell(i, n_attributes - 1)==classValues[actual]) {
						confusion[classification][actual]++;
						break;//we've found the class
					}
				}			
			}
			else if (frfs.LOOCV) {
				System.out.println(frfs.classMap.get(fm.getCell(dec.nextSetBit(0), n_attributes - 1))+"\n"+frfs.classMap.get(data2.getCell(i, n_attributes - 1)));	
				//System.out.println(fm.getCell(dec.nextSetBit(0), n_attributes - 1)+"\n"+data2.getCell(0, n_attributes - 1));		
			}
		}
		if (!frfs.visualise) {
			System.err.println((accuracy / dObjects) * 100);
			if (!frfs.LOOCV) {
				System.out.println((accuracy / dObjects));
			}
			if (frfs.outputConfusionMatrix) {
				System.err.println("Confusion matrix:");
				for (int i=0;i<decision.size();i++) System.err.print("\t"+frfs.classMap.get(classValues[i]));
				System.err.println();

				for (int i=0;i<decision.size();i++) {
					System.err.print(frfs.classMap.get(classValues[i]));
					for (int j=0;j<decision.size();j++) {
						System.err.print("\t"+confusion[i][j]);
					}
					System.err.println();
				}
			}

		} else
			System.out.print(vis);
	}

	private final double[] classifyObjectVQNN(BitSet dec, int i, Matrix data2,
			Subset reduct, ArrayList<Index> objectIndex) {
		double[] ret = new double[2];
		double val = 0;
		double denom = 0;

		// work out VQRS upper and lower approximations

		for (int o = 0; o < frfs.K; o++) {
			double mem = 0;
			index = objectIndex.get(o);

			if (dec.get(index.attribute))
				mem = 1;
			val += FuzzySet.tnorm(index.sim, mem);

			denom += index.sim;
		}

		ret[0] = Q(val / denom, alpha_u, beta_u);// lower
		ret[1] = Q(val / denom, alpha_l, beta_l);// upper
		return ret;
	}

	private final int classify2fnn(int i, Matrix data2, Subset reduct,
			ArrayList<Index> oI) {
		int decisionClass = -4;
		double bestClassification = -1;

		Iterator<BitSet> e = decision.iterator();
		BitSet dec;
		int counter = 0;
		double classification = -1;
		// System.err.println(decision.size());
		// for each decision class
		while (e.hasNext()) {
			dec =  e.next();
			classification = classifyObjectfnn(dec, i, data2, reduct, oI);
			// System.err.print(classification+" ");
			if (classification > bestClassification) {
				bestClassification = classification;
				decisionClass = counter;

			}
			if (frfs.visualise && counter == frfs.decisionClass) {

				// vis+=data2.getCell(i,0)+ " "+data2.getCell(i,1)+" ";
				// vis+=classification[0]+"\n";
				vis += classification;

				if ((i + 1) % frfs.num_vals2 == 0) {
					if (((i + 1) / frfs.num_vals2) < frfs.num_vals1)
						vis += "\n"
								+ frfs.labels_vert[((i + 1) / frfs.num_vals2)]
										+ ", ";
				} else
					vis += ", ";

			}
			counter++;
		}

		return decisionClass;
	}

	double[] oc;

	private final int classify2Ownership(int i, Matrix data2, Subset reduct,
			ArrayList<Index> objectIndex) {
		int decisionClass = -4;
		double bestClassification = -1;

		Iterator<BitSet> e;
		BitSet dec;
		int counter = 0;
		// setup kappa
		double[] kappa = new double[n_attributes - 1];

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			kappa[a] = 0;
			double sum = 0;

			for (int o = 0; o < n_objects; o++) {
				sum += Math.pow(fm.getCell(o, a) - data2.getCell(i, a), 2);
			}
			kappa[a] = n_objects_d / (2 * Math.pow(sum, 1 / (frfs.q - 1)));

		}

		// represents memberships of object i to decision classes
		oc = new double[decision.size()];
		for (int o = 0; o < decision.size(); o++)
			oc[o] = 0;

		for (int o = 0; o < n_objects; o++) {
			double d = 0;

			for (int a = reduct.nextFeature(0); a >= 0; a = reduct
					.nextFeature(a + 1)) {
				d += kappa[a]
						* Math.pow(fm.getCell(o, a) - data2.getCell(i, a), 2);
			}
			objectIndex.add(new Index(o, 1 / d)); // sort according to inverse
			// of distance
		}

		// sort objects according to their similarity to object i
		Collections.sort(objectIndex);

		// for each object in the neighbourhood
		for (int o = 0; o < frfs.K; o++) {
			index = objectIndex.get(o);
			/*
			 * double d=0;
			 * 
			 * for (int a = reduct.nextSetBit(0);a >=0; a =
			 * reduct.nextSetBit(a+1)) {
			 * d+=kappa[a]*Math.pow(fm.getCell(index.attribute,a) -
			 * data2.getCell(i,a),2); }
			 */

			counter = 0;
			e = decision.iterator();
			while (e.hasNext()) {
				dec =  e.next();
				double mem = 0;
				if (dec.get(index.attribute))
					mem = 1;

				// oc[counter] +=
				// (1/(double)frfs2.K)*(mem)*(Math.exp(-Math.pow(d,1/(frfs2.q -
				// 1))));
				oc[counter] += (1 / (double) frfs.K)
						* (mem)
						* (Math.exp(-Math.pow((1 / index.sim), 1 / (frfs.q - 1))));
				counter++;
			}
		}

		for (int o = 0; o < decision.size(); o++) {

			if (oc[o] > bestClassification) {
				bestClassification = oc[o];
				decisionClass = o;
			}
			if (frfs.visualise && o == frfs.decisionClass) {

				// vis+=data2.getCell(i,0)+ " "+data2.getCell(i,1)+" ";
				// vis+=classification[0]+"\n";
				vis += oc[o];

				if ((i + 1) % frfs.num_vals2 == 0) {
					if (((i + 1) / frfs.num_vals2) < frfs.num_vals1)
						vis += "\n"
								+ frfs.labels_vert[((i + 1) / frfs.num_vals2)]
										+ ", ";
				} else
					vis += ", ";

			}
		}
		return decisionClass;
	}

	private final int classify2(int i, Matrix data2, Subset reduct,
			ArrayList<Index> oI) {
		int decisionClass = -1;
		// double bestClassification=-1;
		double[] best = new double[2];
		best[0] = -1f;
		best[1] = -1f;

		Iterator<BitSet> e = decision.iterator();
		BitSet dec;
		int counter = 0;

		// for each decision class
		while (e.hasNext()) {
			dec =  e.next();
			double[] classification = new double[2];
			if (frfs.average)
				classification = classifyObjectAvg(dec, i, data2, reduct, oI);
			else if (frfs.command == frfs.VQNN)
				classification = classifyObjectVQNN(dec, i, data2, reduct, oI);
			else
				classification = classifyObject(dec, i, data2, reduct, oI);
			/*
			 * if (counter==1&&classification>0) System.err.println(i+":
			 * "+counter+" -> "+classification);
			 */

			if (classification[0] > best[0]) {
				if (classification[1] >= best[1]) {
					best[0] = classification[0];
					best[1] = classification[1];
					decisionClass = counter;
				}
			}
			if (frfs.visualise && counter == frfs.decisionClass) {

				// vis+=data2.getCell(i,0)+ " "+data2.getCell(i,1)+" ";
				// vis+=classification[0]+"\n";
				vis += classification[0];

				if ((i + 1) % frfs.num_vals2 == 0) {
					if (((i + 1) / frfs.num_vals2) < frfs.num_vals1)
						vis += "\n"
								+ frfs.labels_vert[((i + 1) / frfs.num_vals2)]
										+ ", ";
				} else
					vis += ", ";

			}
			counter++;
		}

		return decisionClass;
	}

	String vis = "";

	Index index;

	private final double classifyObjectfnn(BitSet dec, int i, Matrix data2,
			Subset reduct, ArrayList<Index> objectIndex) {

		double sum = 0;
		double denom = 0;

		for (int o = 0; o < frfs.K; o++) {
			double mem = 0;
			index = objectIndex.get(o);

			if (dec.get(index.attribute)) {
				mem = 1;
			}

			double val = index.sim;// (double)Math.pow(index.sim,-(q-1));
			// System.err.print(index.sim+" ");
			sum += val * mem;
			denom += val;

		}
		// System.err.println("\n");
		if (denom == 0)
			return 0;
		else
			return sum / denom;
	}

	private final double classifyObjectOwnership(BitSet dec, int i,
			Matrix data2, Subset reduct, ArrayList<Index> objectIndex) {

		double sum = 0;

		for (int o = 0; o < frfs.K; o++) {
			double mem = 0;
			index = objectIndex.get(o);

			if (dec.get(index.attribute)) {
				mem = 1;
			}

			sum += index.sim * mem;

		}

		return sum / n_objects_d;
	}

	private final double[] classifyObject(BitSet dec, int i, Matrix data2,
			Subset reduct, ArrayList<Index> objectIndex) {
		double[] ret = new double[2];
		double lower = 1;
		double upper = 0;
		// System.out.println(dec);

		// work out upper and lower approximations

		for (int o = 0; o < frfs.K; o++) {
			double mem = 0;
			index = objectIndex.get(o);

			if (dec.get(index.attribute))
				mem = 1;
			lower = Math.min(FuzzySet.impl(index.sim, mem), lower);
			upper = Math.max(FuzzySet.tnorm(index.sim, mem), upper);

		}

		ret[0] = lower;// (lower-1)/(double)frfs2.K;
		ret[1] = upper;// /(double)frfs2.K;
		return ret;
	}

	// same as above but uses average instead of inf (Math.min)
	private final double[] classifyObjectAvg(BitSet dec, int i, Matrix data2,
			Subset reduct, ArrayList<Index> objectIndex) {
		double[] ret = new double[2];
		double lower = 0;
		double upper = 0;

		// work out upper and lower approximations
		for (int o = 0; o < frfs.K; o++) {
			double mem = 0;
			index = objectIndex.get(o);

			if (dec.get(index.attribute))
				mem = 1;

			lower += FuzzySet.impl(index.sim, mem);
			upper += FuzzySet.tnorm(index.sim, mem);

		}

		ret[0] = lower / frfs.K;
		ret[1] = upper / frfs.K;
		return ret;
	}

	// need to remove need for tnorm here as it's too constraining, unless a
	// suitable
	// similarity metric is chosen (e.g. Gaussian)
	// best to use Euclidean metric?
	private final double sim(int o, int i, Matrix data2, Subset reduct) {
		double ret = 1;

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			ret = FuzzySet.compose(fuzzySimilarity(a, fm.getCell(o, a), data2
					.getCell(i, a)), ret);

		}

		return ret;
	}

	// fuzzy knn metric
	private final double sim2(int o, int i, Matrix data2, Subset reduct) {
		double ret = 1;

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			ret += Math.pow((fm.getCell(o, a) - data2.getCell(i, a)), 2);
		}
		ret = 1 / Math.pow(ret, 1 / (frfs.q - 1));

		return ret;
	}

	// wang's fuzzy-rough knn metric - Wang et al
	private final double sim3(int o, int i, Matrix data2, Subset reduct) {
		double ret = 1;
		double sum = 0;

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			ret += Math.pow((fm.getCell(o, a) - data2.getCell(i, a)), 2);
			sum += data2.getCell(i, a);
		}
		sum /= n_objects_d;
		double norm = 0;

		for (int a = reduct.nextFeature(0); a >= 0; a = reduct.nextFeature(a + 1)) {
			norm += Math.pow(data2.getCell(i, a) - sum, 2);

		}

		norm /= n_objects_d;

		return Math.exp(-ret / norm);
	}

	double bestGamma;

	double prevGamma = 100;// not used currently

	// using lower approximations/gain ratio/boundary region
	public final Subset quickReduct2(Subset core, double ALPHA) {
		Reduct = new Subset(n_attributes);

		initCurrent();

		System.err.println("-------------------------\nDataset consistency: "
				+ (c_divisor) + " (normalized)");

		// initCurrent();
		bestGamma = 0;
		Reduct = new Subset(n_attributes); // core = new
		// Subset(n_attributes-1);
		// core.set(3);

		// sort out core-related things
		if (core != null && core.cardinality() > 0) {
			Reduct.union(core);
			if (frfs.OPTIMISE != 0) {
				for (int a = core.nextFeature(0); a >= 0; a = core
						.nextFeature(a + 1)) {
					if (core.contains(a)) {
						// clear(temprel);
						generatePartition(current, a, Reduct);
						setCurrent(temprel);
					}
				}
			}
			gamma = evaluate(temprel, -1, Reduct) / c_divisor;
			if (gamma >= 1) {
				System.err.println(Reduct + " => 1.0");
				Reduct.add(n_attributes - 1);
				System.err.println("Final subset is " + Reduct + " |"+ Reduct.cardinality() + "|");
				return Reduct;
			}
			bestGamma = gamma;
			System.err.println(Reduct + " => " + gamma);
		}

		// if ALPHA is zero or hasn't been set, then set it to 1
		if (ALPHA == 0)
			ALPHA = 1;

		boolean go = true;
		double gamma = -Double.MAX_VALUE;

		int bestAttr = 0;
		double oldGamma = -Double.MAX_VALUE;
		double bestEntropy = 10000;
		prevEntropy = log2(decCardinality);
		String reducts = "Reducts found: ";

		boolean noPartitions = ((frfs.command == frfs.CORNELIS)
				|| (frfs.command == frfs.INTERVAL) || (frfs.command == frfs.FDM));

		while (go) {
			// augment this with nextClearBit etc to speed up
			for (int a = 0; a < n_attributes - 1; a++) {
				if (!Reduct.contains(a)) {
					Reduct.add(a);
					// clear(temprel);

					if (!noPartitions)
						generatePartition(current, a, Reduct);

					gamma = evaluate(temprel, a, Reduct);
					gamma /= c_divisor;

					if (gamma >= bestGamma) {
						if (gamma > 1)
							gamma = 1;
						bestEntropy = entropy;
						bestGamma = gamma;
						bestAttr = a;
						if (gamma == 1)
							reducts += Reduct.toString() + " - ";
					}
					if (verbose) {
						System.err.println(Reduct + ": " + gamma);
					}
					Reduct.remove(a);
				}
			}
			Reduct.add(bestAttr);
			prevEntropy = bestEntropy;

			if (bestGamma >= ALPHA) {
				go = false;
			}
			if ((frfs.OPTIMISE == 1 || frfs.OPTIMISE == 2) && !noPartitions) {
				generatePartition(current, bestAttr, Reduct);
				setCurrent(temprel);
			}
			System.err.println(Reduct + " => " + bestGamma);
			if (frfs.command == frfs.VQRS || frfs.command == frfs.VQRS2) {
				bestGamma = 0;
			} // because of nonmonotonicity
			else if (oldGamma >= bestGamma) {
			}// go=false;

			oldGamma = bestGamma;
			if (Reduct.cardinality() == n_attributes - 1)
				break;

		}

		Reduct.add(n_attributes - 1);
		System.err.println("Final subset is ");
		System.out.print(Reduct+"\t");
		System.err.print("|");
		System.out.print(Reduct.cardinality()+"\t");
		System.err.println("|");
		Reduct.remove(n_attributes - 1);
		if (frfs.command == frfs.VQRS || frfs.command == frfs.VQRS2) {
			System.err.println("Dependency is " + (calculateGamma(current, -1)/dependencyFull));
		}
		Reduct.add(n_attributes - 1);

		System.err.println();
		System.err.println(reducts);
		return Reduct;
	}

	private double evaluate(Relation rel, int a, Subset reduct) {
		double gam = 0;
		Reduct = reduct;

		switch (frfs.command) {
		case frfs.ENTROPY_REDUCT:
			gam = calculateEntropy(rel, a);
			break;
		case frfs.ORIGINAL_FRFS:
			gam = calculateGamma4(rel, a);
			break;
		case frfs.BOUNDARY_ENTROPY:
			gam = calculateBoundaryEnt(rel, a);
			break;
		case frfs.BOUNDARY:
			gam = calculateBoundary(rel, a);
			break;
		case frfs.INTERVAL_BOUNDARY:
			gam = calculateBoundaryT2(reduct);
			break;
		case frfs.GAIN_RATIO:
			gam = calculateGR(rel, a);
			break;
		case frfs.DELTA:
			gam = calculateDelta(rel, a);
			break;
		case frfs.WEAK_DELTA:
			gam = calculateDelta2(rel, a);
			break;
		case frfs.CORNELIS:
			gam = alphaValue(reduct);
			break;
		case frfs.FDM:
			gam = fdm2(reduct);
			break;
		case frfs.VQRS:
			gam = calculateVQRS(rel);
			break;
		case frfs.VQRS2:
			gam = calculateVQRS2(rel);
			break;
		case frfs.INTERVAL_FDM:
			gam = fdmT2(reduct);
			break;
		case frfs.INTERVAL:
			gam = calculateGammaT2(reduct);
			break;
		case frfs.WEAK_LOWER:
			gam = calculateGamma2(rel, a);
			break;
		default:
			gam = calculateGamma(rel, a);
			break;
		}

		return gam;
	}

	private final void setCurrent(Relation rel) {
		if (frfs.OPTIMISE!=4) {
			for (int i = 0; i < n_objects; i++) {
				for (int j = i; j < n_objects; j++) {
					current.setCell(i, j, rel.getCell(i, j));
					current.setCell(j, i, current.getCell(i, j));
				}
			}
		}
	}

	// FRFS-SAT - determine fuzzy clauses
	public void fuzzyDiscernibility() {
		constructSimilarity(n_attributes);
		for (int o = 0; o < n_objects; o++) {
			for (int o1 = o + 1; o1 < n_objects; o1++) {

				for (int a = 0; a < n_attributes - 1; a++) {
					System.out.print((1 - relations[a].getCell(o, o1)) + " ");
				}

				System.out.println(1 - relations[n_attributes - 1].getCell(o,
						o1));

			}
		}
	}

	// forces lower approx
	public void isReduct(Subset reduct) {
		reduct.remove(n_attributes - 1);

		if (frfs.OPTIMISE >= 2) {
			constructSimilarity(n_attributes);
		}
		bestGamma = Double.MIN_VALUE;
		frfs.command = frfs.LOWER;

		setConsistency();

		generatePartition(reduct);
		double gamma = calculateGamma(temprel, -1);
		System.err.println("Dataset consistency: " + c_divisor);
		System.err.println(reduct + " => " + gamma / c_divisor);

	}

	// sets c_divisor (the consistency of the entire dataset)
	public void setConsistency() {
		Subset full = new Subset(n_attributes);
		for (int a = 0; a < n_attributes - 1; a++)
			full.add(a);

		c_divisor = consistency(full);

		if (c_divisor == 0 || c_divisor == Double.NaN) {
			System.err.println("\n*** Inconsistent data (full dataset value = "
					+ c_divisor + " for this measure) ***\n");
			System.exit(1);
		}

		if (frfs.command==frfs.VQRS||frfs.command==frfs.VQRS2) {
			int command = frfs.command;
			frfs.command = frfs.LOWER;
			dependencyFull = consistency(full);
			frfs.command = command;
		}
	}

	// return the evaluation of the full dataset
	public double consistency(Subset full) {
		bestGamma = -1;
		double ret = 0;

		if (frfs.OPTIMISE >= 2) {
			initCurrent();

			for (int a = 0; a < n_attributes - 1; a++) {
				generatePartition2(current, a);
				setCurrent(temprel);
			}
		} else {
			generatePartition(full);
			setCurrent(temprel);
		}

		ret = evaluate(current, -1, full);

		// if (ret==0) ret=1;

		return ret;
	}

	/**
	 * Works out the core of the dataset
	 * this is based on the lower approximation only (however, a similar concept could be used for any measure)
	 * @return the core as a Subset
	 **/
	public Subset computeCore() {
		Subset full = new Subset(n_attributes - 1);
		Subset core = new Subset(n_attributes - 1);
		for (int a = 0; a < n_attributes - 1; a++)
			full.add(a);
		bestGamma = Double.MIN_VALUE;
		double gammaFull = -1;
		int temp = frfs.command;
		frfs.command = frfs.LOWER;
		gammaFull = consistency(full);
		frfs.command = temp;

		System.err.println("Calculate core attributes");
		System.err.println("Full dataset dependency is: " + gammaFull);
		double gamma = 0;
		System.err.print("Core: [ ");

		if (frfs.OPTIMISE == 2 || frfs.OPTIMISE == 3) {

			for (int a1 = 0; a1 < n_attributes - 1; a1++) {
				initCurrent();
				// construct partition
				for (int a = 0; a < n_attributes - 1; a++) {
					if (a != a1) {
						generatePartition2(current, a);
						setCurrent(temprel);
					}
				}
				gamma = calculateGamma(current, -1);
				if (gamma < gammaFull) {
					core.add(a1);
					System.err.print(a1 + " ");
				}
			}

		} else {
			for (int a = 0; a < n_attributes - 1; a++) {
				full.remove(a);
				if (frfs.OPTIMISE!=4) generatePartition(full);
				gamma = calculateGamma(temprel, -1);
				if (gamma < gammaFull) {
					core.add(a);
					System.err.print(a + " ");
				}
				full.add(a);
			}
		}
		System.err.println("]");
		return core;
	}

	/**
	 * Works out the core of the dataset, and uses this as the starting point for QuickReduct
	 * this is based on the lower approximation only (however, a similar concept could be used for any measure)
	 * 
	 * @param ALPHA, the degree of reducthood (default = 1)
	 * @return the final reduct as a Subset
	 **/
	public Subset coreReduct(double ALPHA) {
		Subset core = computeCore();
		Subset finalReduct = null;

		finalReduct = quickReduct2(core, ALPHA);

		return finalReduct;
	}

	/**
	 * Evaluates a subset via FDM
	 * 
	 * @param reduct, the subset to consider
	 * @return double, the evaluation
	 **/
	private double fdm2(Subset reduct) {
		double d1, d2, difd;
		double a1, a2, disc, d;
		double ret;

		if (!frfs.average)
			ret = 1; // min
		else
			ret = 0;

		double valueRA = 0;

		for (int i = 0; i < n_objects - 1; i++) {
			for (int j = i + 1; j < n_objects; j++) {

				// decision attribute values
				d1 = fm.getCell(i, n_attributes - 1);
				d2 = fm.getCell(j, n_attributes - 1);
				difd = 1.00 - fuzzySimilarity(n_attributes - 1, d1, d2);

				//valueRA = 0;

				valueRA=1;

				// optimization
				//if (difd != 0) {
				// conditional attributes in reduct
				for (int a = reduct.nextFeature(0); a >= 0; a = reduct
						.nextFeature(a + 1)) {
					a1 = fm.getCell(i, a);
					a2 = fm.getCell(j, a);

					disc = 1 - fuzzySimilarity(a, a1, a2);

					// s-norm
					//valueRA = FuzzySet.compose2(disc, valueRA); // determined by -comp
					valueRA = FuzzySet.compose(1-disc,valueRA);
					//if (d==3&&x==1) System.err.println(y+" => "+currLower);

					//if (valueRA == 1)	break;
				}
				//d = FuzzySet.impl(difd, valueRA);
				d = FuzzySet.impl(valueRA,1-difd);
				//} else
				//d = 1;


				if (!frfs.average) {
					ret = FuzzySet.tnorm(ret, d);
					// ret = Math.min(ret,d); - may produce non-reducts
					//if (ret == 0) return 0;
				} else
					ret += d;

				//System.err.println(d+" => "+ret);
			}
		}
		if (frfs.average)
			return ret / ((n_objects_d * n_objects_d - n_objects_d) / 2); // normalise
		else
			return ret;
	}

	private double alphaValue(Subset reduct) {
		double xi, xj, difd;
		double oij;
		double ret;

		if (!frfs.average)
			ret = 1; // min
		else
			ret = 0;

		for (int i = 0; i < n_objects - 1; i++) {
			for (int j = i + 1; j < n_objects; j++) {

				// decision attribute values
				xi = fm.getCell(i, n_attributes - 1);
				xj = fm.getCell(j, n_attributes - 1);

				// 1 - R_d(x_i,x_j)
				difd = 1.00 - fuzzySimilarity(n_attributes - 1, xi, xj);

				double valueRA = 1;
				double s = 0;
				double saux;

				// optimization for qualitative decisions I(0,x)=1
				if (difd != 0) {

					// conditional attributes
					for (int a = 0; a < n_attributes - 1; a++) {
						xi = fm.getCell(i, a);
						xj = fm.getCell(j, a);

						int astar = 0;
						if (reduct.contains(a))
							astar = 1;

						oij = 1 - fuzzySimilarity(a, xi, xj);

						// numerator - S(O*_ij(a*)a*,...,)
						s = FuzzySet.snorm(oij * astar, s);

						// denominator - keeping track of R_A(x_i,x_j)
						valueRA = FuzzySet.compose(1 - oij, valueRA);
					}
					saux = FuzzySet.impl(difd, s ); //remove local normalization: / (1 - valueRA)
				} else
					saux = 1;

				if (!frfs.average)
					ret = Math.min(ret, saux);
				else
					ret += saux;

			}
		}
		if (frfs.average)
			return ret / ((n_objects_d * n_objects_d - n_objects_d) / 2); // normalise
		else
			return ret;
	}

	// using lower approximations/gain ratio etc
	public final Subset reverseReduct(double val) {
		Reduct = new Subset(n_attributes);
		boolean go = true;
		double gamma = -Double.MAX_VALUE;
		bestGamma = n_objects;
		int bestAttr = 0;
		// double divisor = n_objects_d;
		for (int a = 0; a < n_attributes - 1; a++)
			Reduct.add(a);
		double full;

		if (val == 0) { // i.e. no value given for reduct quality
			full = consistency(Reduct);
			bestGamma = full;
			System.err.println("Dataset consistency: " + (full));
		} else
			full = val;

		bestGamma = full;

		System.err.println("Dataset consistency: " + (c_divisor)
				+ " (normalized)");
		String reducts = "Reducts found: ";
		boolean noPartitions = ((frfs.command == frfs.CORNELIS)
				|| (frfs.command == frfs.INTERVAL) || (frfs.command == frfs.FDM));

		while (go) {
			bestAttr = -1;
			bestGamma = full;
			// augment this with nextClearBit etc to speed up
			//for (int a = 0; a < n_attributes - 1; a++) {
			for (int a=n_attributes-2;a>=0;a--) {
				if (Reduct.contains(a)) {
					Reduct.remove(a);

					if (!noPartitions)
						generatePartition(null, -1, Reduct);

					gamma = evaluate(temprel, a, Reduct);
					gamma /= c_divisor;

					if (gamma>=bestGamma) {

					}
					else
						Reduct.add(a);
				}
			}
			if (bestAttr == -1) {
				go = false;
				break;
			}

			Reduct.remove(bestAttr);
			System.err.println(Reduct + " => " + bestGamma);
			if (Reduct.cardinality() == 1) {
				break;
			}
		}
		Reduct.add(n_attributes - 1);
		System.err.println("Final subset is " + Reduct + " |"
				+ Reduct.cardinality() + "|");

		Reduct.remove(n_attributes - 1);


		if (frfs.command == frfs.VQRS || frfs.command == frfs.VQRS2) {
			System.err.println("Dependency is " + (calculateGamma(current, -1)/dependencyFull));
		}
		Reduct.add(n_attributes - 1);

		System.err.println();
		System.err.println(reducts);

		return Reduct;
	}

	//used for ACO, GA etc
	public double calculate(Subset reduct) {
		double ret = 0;

		generatePartition3(reduct);
		setCurrent(temprel);
	
		ret = evaluate(current, -1, reduct);

		return ret/c_divisor;
	}
}