package search;
/*
 * This models an individual ant in the feature selection process
 * Each ant has a note whether it has finished its journey, an 
 * identifying number, a reduct (the one it is constructing), a note
 * of the reduct's gamma value and finally a record of the path taken.
 */


import set.Subset;
import java.util.ArrayList;

public final class Ant {
	boolean finished=false; //has the ant finished construction?
	int number; //identifier
	Subset reduct; // set of attributes this ant has chosen so far
	double heur=-1; //gamma so far
	ArrayList<Edge> edges; //the list of edges that make up this path

	//create an ant, with identifier n
	public Ant(int n) {number = n;edges = new ArrayList<Edge>();}

	//create an ant, with identifier n
	// 'initial' is a single attribute that we want our initial
	// reduct to contain.
	// 'attributes' is the number of attributes in the dataset - this
	// includes the decision attribute, so we use attributes-1
	public Ant(int initial, int n,int attributes) {
		reduct = new Subset(attributes-1); //create an initially empty reduct
		reduct.add(initial); //set the individual attribute
		number = n;
		edges = new ArrayList<Edge>();
		//the first part of this path is the attribute 'initial'
		edges.add(new Edge(initial)); 
	}

	//set the attribute 'index' in the ant's reduct
	public void set(int index) {
		reduct.add(index);
	}

	//remove the attribute 'index' in the ant's reduct
	public void clear(int index) {
		reduct.remove(index);
	}

	//add an edge to the ant's path
	public void addEdge(Edge e) {
		edges.add(e);
	}
}

