package search;

//TODO - store all chromosomes in a hash table. when we create a new
//generation, look up to see if they've been calculated before (need to
//store both the bitset and the fitness together in one class though)

import java.util.*;

import frfs2.RSys2;
import set.Subset;

public final class GeneticSearch {
	int attributes;
	int objects;
	Hashtable<Subset, Subset> ht; // contains reducts
	Hashtable<Subset, Chromosome> pop;
	boolean verbose;
	BitSet temp = null;
	BitSet empty = null;
	BitSet remaining = null;
	Enumeration<Subset> reds;
	Chromosome c = null;
	double[] gammas;
	double m_sumFitness = 0;
	Random random = new Random();
	Subset[] population;
	Subset individual, small;
	double[] fitness;
	int generations = -1;
	double mutation = -1;
	double crossover = -1;
	int best = -1;
	Subset[] newPop = null;
	double gam = 0;
	double quality = 1;
	RSys2 rsys2;

	public GeneticSearch(RSys2 r, int o, int a, boolean v, int p, int g,
			double m, double c, double q) {
		rsys2 = r;
		rsys2.setConsistency();

		objects = o;
		attributes = a;
		verbose = v;
		population = new Subset[p];
		fitness = new double[p]; // fitness of corresponding chromosomes
		gammas = new double[p];
		generations = g;
		mutation = m;
		crossover = c;
		quality = q; // reduct quality (only used for Cornelis approach)
		temp = new BitSet(objects);
		ht = new Hashtable<Subset, Subset>();
		pop = new Hashtable<Subset, Chromosome>();
	}

	private float randomHeuristic() {
		return (float) (Math.abs(random.nextInt()) % (attributes)) / attributes;
	}

	private final Subset randomReduct() {
		int attr;
		Subset ret = new Subset(attributes);
		int no_of_attrs = Math.abs(random.nextInt()) % (attributes - 1);
		if (no_of_attrs < 1)
			no_of_attrs = 3;
		if (no_of_attrs > 19)
			no_of_attrs = 19;

		for (int i = 0; i < no_of_attrs; i++) {
			attr = Math.abs(random.nextInt()) % (attributes - 1);
			ret.add(attr);
		}

		return ret;
	}

	// initialise the population
	public int initPopulation() {
		int t = -1;
		double bestFitness = -1;

		for (int i = 0; i < population.length; i++) {
			population[i] = randomReduct();
			fitness[i] = evaluate(i);
			if (fitness[i] > bestFitness) {
				bestFitness = fitness[i];
				t = i;
			}
		}
		return t;
	}

	private void evaluatePopulation() {
		int best = 0;
		pop = new Hashtable<Subset, Chromosome>(population.length);

		for (int i = 0; i < population.length; i++) {
			if (!pop.containsKey(population[i])) {
				fitness[i] = evaluate(i);
				pop
						.put(population[i], new Chromosome(population[i],
								gammas[i]));
			} else {
				c = pop.get(population[i]);
				fitness[i] = c.gamma;
			}
			if (fitness[i] > fitness[best])
				best = i;
		}
	}

	private void add(int i) {
		if (!ht.containsKey(population[i]))
			ht.put(population[i], population[i]);
	}

	// evaluation function
	private double evaluate(int i) {
		if (population[i] != null && population[i].cardinality() > 0) {
			gam = rsys2.calculate(population[i]);
			gammas[i] = gam;
			if (gam >= quality) { // subset is good enough
				add(i);
				if (verbose)
					System.err.println(population[i] + " |"
							+ population[i].cardinality() + "|  " + gam);
			}
			return ((0.5 * gam) + (0.5 * (((double) (attributes - population[i]
					.cardinality())) / (double) attributes)));
		} else {
			gammas[i] = 0;
			return 0;
		}
	}

	private boolean converged() {
		boolean ret = false;
		/*
		 * float max=-1; float min=2; float curr=-1;// = fitness[0];
		 * 
		 * for (int i=1;i<population.length;i++) { curr = fitness[i]; if (curr<min)
		 * min=curr; if (curr>max) max=curr; } ret=(max==min);
		 */
		return ret;
	}

	private int select() {

		int i;
		double r, partsum;

		partsum = 0;
		r = random.nextDouble() * m_sumFitness;
		for (i = 0; i < population.length; i++) {
			partsum += fitness[i];
			if (partsum >= r) {
				break;
			}
		}
		if (i == population.length)
			i--;
		return i;
	}

	// perform selection, crossover and mutation
	private void generation() {

		for (int i = 0; i < population.length; i++) {
			// newPop[i] = (BitSet)population[i].clone();
			m_sumFitness += fitness[i];
		}
		int parent1 = -1;
		int parent2 = -1;
		double r = -1f;

		newPop[0] = (Subset) population[best].clone();
		newPop[1] = (Subset) population[0].clone();

		for (int i = 2; i < population.length; i += 2) {
			parent1 = select();
			parent2 = select();

			newPop[i] = (Subset) population[parent1].clone();
			newPop[i + 1] = (Subset) population[parent2].clone();
			// System.out.println("crossover");
			if (parent1 == parent2) {
				int s = Math.abs(random.nextInt()) % (attributes - 1);
				if (newPop[i].contains(s))
					newPop[i].remove(s);
				else
					newPop[i].add(s);
			} else { // crossover
				r = random.nextDouble();
				if (r < crossover) {
					int cp = Math.abs(random.nextInt());
					cp %= (attributes - 3);
					cp++;

					for (int a = 0; a < cp; a++) {
						if (population[parent1].contains(a))
							newPop[i + 1].add(a);
						else
							newPop[i + 1].remove(i);

						if (population[parent2].contains(a))
							newPop[i].add(a);
						else
							newPop[i].remove(i);
					}

				}
			}

			// mutation
			for (int k = 0; k < 2; k++) {
				for (int a = 0; a < attributes - 1; a++) {
					r = random.nextDouble();
					if (r < mutation) {
						if (newPop[i + k].contains(a))
							newPop[i + k].remove(a);
						else
							newPop[i + k].add(a);
					}
				}
			}
			// fitness[i] = evaluate(i);

		}
		// System.arraycopy(newPop,0,population,0,newPop.length);
		population = newPop;

	}

	public Subset run() {
		temp = new BitSet(objects);

		newPop = new Subset[population.length];
		best = initPopulation();
		// evaluatePopulation();
		// reportPopulation(); //possibly just report the best

		for (int it = 0; it < generations; it++) {
			generation();
			evaluatePopulation();

			// if (verbose) System.out.println(population[best] + " ==> " +
			// fitness[best]);
			if (converged()) {
				System.err.println("...converged");
				break;
			}

		}
		// System.out.println("Final best is "+ population[best] + " ==> " +
		// fitness[best]);
		// System.out.print("Reducts found are: ");

		reds = ht.elements();
		int smallest = attributes - 1;

		while (reds.hasMoreElements()) {
			individual =  reds.nextElement();
			int s = individual.cardinality();
			if (s == 0) {
			} else if (s < smallest) {
				smallest = s;
				small = (Subset) individual.clone();
			}
		}
		if (small != null && small.cardinality() > 0) {
			small.add(attributes - 1);
			smallest = small.cardinality();
		}
		System.err.println();
		System.err.println("Smallest reduct is: " + small + " |" + smallest
				+ "|");
		return small;
	}
}

// 2*a*fitness+a-size
// a*gam

class Chromosome {
	Subset string;
	double gamma;

	public Chromosome(Subset s, double g) {
		string = s;
		gamma = g;
	}
}
