package search;

import java.util.StringTokenizer;
import java.io.*;
import java.util.*;

import set.FuzzySet;

import frfs2.Matrix;

public class Search {
	static int cols=0;
	static int rows=0;
	static double current=0;
	static boolean verbose=false;
	static int CONNECT=2;
	static boolean SUM=true; //use sum of clause satisfactions
	static float MAX;

	public static void main(String[] args) throws Exception {
		BufferedReader in = new BufferedReader(new FileReader(args[0]));
		String line = in.readLine();
		StringTokenizer st = new StringTokenizer(line," ,");

		if (args.length>1) {
			if (args[1].equals("-v")) verbose=true;
			else if (args[1].equals("-c")) {
				CONNECT = Integer.parseInt(args[2]);
			} 
			if (args[3]!=null) {
				if (args[3].equalsIgnoreCase("-sum")) SUM=true;
				else if (args[3].equalsIgnoreCase("-tnorm")) SUM=false;
			}
		}
		while(st.hasMoreTokens()) {
			st.nextToken();
			cols++;
		}

		while(line!=null&&line.length()>0) {
			rows++;
			line = in.readLine();
		}

		in = new BufferedReader(new FileReader(args[0]));
		Matrix data = new Matrix(rows,cols);

		line = in.readLine();
		String tok;

		for (int i=0;i<rows;i++) {
			st = new StringTokenizer(line," ,");
			for (int j=0;j<cols;j++) {
				tok = st.nextToken();
				double val;
				if (tok.equals("?")) val=0;
				else val = Double.valueOf(tok).doubleValue();
				data.setCell(i,j,val);
			}
			line = in.readLine();
			if (line==null) break;
		}
		in.close();
		System.out.println("Number of clauses: "+rows);
		long start = System.currentTimeMillis(); 

		if (SUM) MAX=rows;
		else MAX=1.0f;

		bf(data);
		float time = (float) (System.currentTimeMillis() - start);
		System.err.println("Time taken = "+(time/1000)+" secs");
	}
	static boolean[] eliminated ;

	public static void bf(Matrix data) {
		BitSet reduct = new BitSet(data.numCols());

		int best=-1;
		double bestval = -1;
		double val1=-1;
		double val=-1;
		eliminated = new boolean[rows];

		for (int o=0;o<rows;o++) eliminated[o]=false;

		while(reduct.cardinality()<data.numCols()-1) {
			best=-1;
			
			for (int a=0;a<cols-1;a++) {
				if (!reduct.get(a)) {
					reduct.set(a);
					if (SUM) val1 = getSat(reduct,data,1.0f);
					else val1 = getSat2(reduct,data,1.0f);


					val = val1;
					if (verbose) System.out.println("  "+reduct+" ==> "+val);
					if (val>=bestval) {bestval=val;best=a;}	    
					reduct.clear(a);
				}
			}
			if (best==-1) {reduct.set(cols-1);System.out.println("Fuzzy Reduct is "+reduct+" |"+reduct.cardinality()+"|");break;}
			reduct.set(best);
			if (bestval==MAX) {System.out.println("Choosing "+reduct+" ==> "+bestval);reduct.set(cols-1);System.out.println("Fuzzy Reduct is "+reduct+" |"+reduct.cardinality()+"|");break;}
			else {
				//eliminate(data,best);
				System.out.println("Choosing "+reduct+" ==> "+bestval);
			}
		}
	}

	//eliminate clauses that contain this attribute
	public static void eliminate(Matrix data, int a) {
		for (int o=0;o<rows;o++) {
			if (data.getCell(o,a)>0) {
				eliminated[o]=true;
			}
			else eliminated[o]=false;
		}
	}

	static double sum;

	//uses the sum of fuzzy clause satisfactions (as in TFS paper)
	public static double getSat(BitSet red, Matrix data,float dec) {
		sum=0;
		FuzzySet.CONNECT = CONNECT;

		for (int o=0;o<rows;o++) {

			current=0;

			for (int a = red.nextSetBit(0);a >=0; a = red.nextSetBit(a+1)) {

				current = FuzzySet.snorm(current,data.getCell(o,a));

			}
			current = FuzzySet.impl(data.getCell(o,cols-1),current);

			sum+=current;
			//sum = FuzzySet.tnorm(sum,current);
		}
		return sum;
	}

	//uses t-norm based approach (more in-keeping with CNF)
	public static double getSat2(BitSet red, Matrix data,float dec) {
		sum=1;
		FuzzySet.CONNECT = CONNECT;

		for (int o=0;o<rows;o++) {

			current=0;

			for (int a = red.nextSetBit(0);a >=0; a = red.nextSetBit(a+1)) {

				current = FuzzySet.snorm(current,data.getCell(o,a));

			}
			current = FuzzySet.impl(data.getCell(o,cols-1),current);


			sum = FuzzySet.tnorm(sum,current);
		}
		return sum;
	}

}
