package set;
//this version changed from objects to FuzzySets
import java.util.*;

import frfs2.Matrix;


public final class FuzzyList {
  SetNode first,last,current;
  int current_idx,count;
  FuzzySet ret;

  public FuzzyList() {
     first = null;
     last = null;
     current = null;
     current_idx = 0;
     count = 0;
  }

  public FuzzyList(FuzzyList sl) {
    if(sl.length() == 0) {
      first = null;
      last = null;
      current = null;
      current_idx = 0;
      count = 0;
    } 
    else {
      SetNode ptr = sl.first;
      first = last = new SetNode(ptr.hd);
      count = 1;

      while( ptr.tl != null ) {
    ptr = ptr.tl;
    last.tl = new SetNode(ptr.hd);
    last = last.tl;
    count++;
      }
      current = first;
      current_idx = 0;
    }
  }

  public final int length() {
    return count;
  }
 
  public int size() {
    return count;
  }

  public boolean isEmpty() {
    return count==0;
  }

  public final Object elementAt(int index) {
    if(index >= count)
      System.err.println("FuzzyList.elementAt, invalid index");

    if(index < current_idx) {
      current = first;
      current_idx = 0;
    }

    for( ; current_idx<index; current_idx++ )
      current = current.tl;

    return current.hd;//.clone();    
  }

  public final void setFirst() {
   current = first;
  }

  public final FuzzySet getNext() {
    ret = current.hd;
    current = current.tl;
    return ret;
  }

  public final float getMembership(Matrix fm, int object) {
    ret = current.hd;
    current = current.tl;
    return ret.getMembership(fm.getCell(object,ret.i));
  }

    public final float getFullMembership(Matrix fm,int object) {
	float temp=1;
	setFirst();
	
	for (int i=0;i<count;i++) {
	    ret = current.hd;
	    current = current.tl;
	    temp = FuzzySet.tnorm(ret.getMembership(fm.getCell(object,ret.i)),temp);
	    if (temp==0) break;
	}
	return temp;
    }

  public final float getMaxMembership(Matrix fm,int object) {
    float temp=0;
    setFirst();

    for (int i=0;i<count;i++) {
      ret = current.hd;
      current = current.tl;
      temp = FuzzySet.snorm(ret.getMembership(fm.getCell(object,ret.i)),temp);
      if (temp==1) break;
    }
    return temp;
  }

  public final void add(FuzzySet s) {
   if( first == null ) {
     first = new SetNode(s);
     last = first;
     current = first;
     current_idx = 0;
   } 
   else {
     last.tl = new SetNode(s);
     last = last.tl;
   }
   count++; 
  }


  public final void del_front(int num) {
    if(num > count)
      System.err.println("FuzzyList.del_front, attempt to delete too many items");
    while(num-->0) {
      current = first.tl;
      first = current;
      count--;
    }
    current = first;
    current_idx = 0;    
  }

  public final void clear() {
    first = null;
    last = null;
    current = null;
    current_idx = 0;
    count = 0;    
  }

  public final void xintersection(FuzzyList b) {
    int a_len = length();
    int b_len = b.length();
    /*   FuzzySet temp=null;

    for(int i=0; i<a_len; i++ ) {
      for( int j=0; j<b_len; j++ ) {  
    temp = elementAt(i);
    temp.and(b.elementAt(j));
    if( temp.length()>0 ) // >1
      add(temp);
      }
    }
    del_front(a_len);*/
  }

  //operator=
  public final FuzzyList equal(FuzzyList sl) {
    clear();
    if( sl.length() == 0 ) {
      first = null;
      last = null;
      current = null;
      current_idx = 0;
      count = 0;
    } 
    else {
      SetNode ptr = sl.first;
      first = last = new SetNode(ptr.hd);
      count = 1;

      while( ptr.tl != null ) {
    ptr = ptr.tl;
    last.tl = new SetNode(ptr.hd);
    last = last.tl;
    count++;
      }
      current = first;
      current_idx = 0;
    }
    return this;    
  }

  public final String toString() {
    String s="{";
    SetNode c = first;

    StringBuffer res = new StringBuffer(s);

    while(c!=null) {
      res.append(c.hd.toString());
      c = c.tl;
    }
    res.append("}");
    return res.toString();
  }

  public FuzzyList clone1() {
    FuzzyList ret = new FuzzyList();
    SetNode ptr = first;
    ret.first = ret.last = new SetNode(ptr.hd);
    ret.count = 1;
    
    while( ptr.tl != null ) {
      ptr = ptr.tl;
      ret.last.tl = new SetNode(ptr.hd);
      ret.last = ret.last.tl;
      ret.count++;
    }
    ret.current = first;
    ret.current_idx = 0;
    
    return ret;
  }
}

final class SetNode {
  SetNode tl;
  FuzzySet hd;

  public SetNode(FuzzySet s) {
    hd = s;
    tl = null;
  }
}

