package set;

//import java.util.*;

public final class FuzzySet {
	boolean discrete;
	boolean empty = false;
	int i = 0;// column number
	int TYPE = -1;

	// experimental: alpha cuts
	static float alphaCut = 0.7f;

	// continuous
	public static final int TRIANGULAR = 1;
	public static final int TRAPEZOIDAL = 2;
	public static final int L_SHOULDERED = 3;
	public static final int R_SHOULDERED = 4;
	public static final int GAUSSIAN = 5;
	public static final int INDIVIDUAL = 6;

	public static int CONNECT = 2;
	public static int COMPOSE = 2;

	public static final int MINMAX = 1;
	public static final int LIMITED = 2;// Lukasiewicz
	public static final int ALGEBRAIC = 3;
	public static final int DRASTIC = 4;
	public static final int TM = 5;

	String name = "";
	float a, b, c, d;
	float m1, m2;
	float scale = 1;
	static float result = 0;

	public FuzzySet(String n, int num, float a1, float b1, float c1, float d1,
			int type) {
		i = num;
		discrete = false;
		name = n;
		TYPE = type;
		a = a1;
		b = b1;
		c = c1;
		d = d1;
	}

	public FuzzySet(String na) {
		discrete = true;
		name = na;
	}

	// fuzzy implication
	public static final float impl(float mufi, float val) {
		switch (CONNECT) {
		case MINMAX:
			result = Math.max(1 - mufi, val);
			break;
		case LIMITED:
			result = Math.min(1, 1 - mufi + val);
			break;
		// S-implicator: case ALGEBRAIC: result = 1 - mufi + (mufi*val);break;
		case ALGEBRAIC:
			if (mufi <= val) {
				result = 1;
			} else {
				result = val / mufi;
			}
			break;
		case TM:
			if (mufi <= val) {
				result = 1;
			} else {
				result = val;
			}
			break;
		default:
			result = Math.max(1 - mufi, val);
			break;
		}
		return result;
	}

	// fuzzy implication
	public static final double impl(double mufi, double val) {
		double result = 0;
		switch (CONNECT) {
		case MINMAX:
			result = Math.max(1.0 - mufi, val);
			break;
		case LIMITED:
			result = Math.min(1.0, 1.0 - mufi + val);
			break;
		// S-implicator: case ALGEBRAIC: result = 1 - mufi + (mufi*val);break;
		case ALGEBRAIC:
			if (mufi <= val) {
				result = 1;
			} else {
				result = val / mufi;
			}
			break;
		case TM:
			if (mufi <= val) {
				result = 1;
			} else {
				result = val;
			}
			break;
		default:
			result = Math.max(1.0 - mufi, val);
			break;
		}
		return result;
	}

	public static final float tnorm(float x, float y) {
		switch (CONNECT) {
		case MINMAX:
			result = Math.min(x, y);
			break;
		case LIMITED:
			result = Math.max(0, x + y - 1);
			break;
		case ALGEBRAIC:
			result = x * y;
			break;
		case TM:
			result = Math.min(x, y);
			break;
		case DRASTIC:
			if (Math.max(x, y) == 1)
				result = Math.min(x, y);
			else
				result = 0;
		}
		// if (result<alphaCut) result=0;
		return result;
	}

	public static final double tnorm(double x, double y) {
		double result = 0;
		switch (CONNECT) {
		case MINMAX:
			result = Math.min(x, y);
			break;
		case LIMITED:
			result = Math.max(0, x + y - 1);
			break;
		case ALGEBRAIC:
			result = x * y;
			break;
		case TM:
			result = Math.min(x, y);
			break;
		case DRASTIC:
			if (Math.max(x, y) == 1)
				result = Math.min(x, y);
			else
				result = 0;
		}
		// if (result<alphaCut) result=0;
		return result;
	}

	public static final float compose(float x, float y) {
		return (float) compose((double) x, (double) y);
	}

	public static final double compose(double x, double y) {
		double result = 0;
		switch (COMPOSE) {
		case MINMAX:
			result = Math.min(x, y);
			break;
		case LIMITED:
			result = Math.max(0, x + y - 1);
			break;
		case ALGEBRAIC:
			result = x * y;
			break;
		case TM:
			result = Math.min(x, y);
			break;
		case DRASTIC:
			if (Math.max(x, y) == 1)
				result = Math.min(x, y);
			else
				result = 0;
		}
		return result;
	}

	public static final float snorm(float x, float y) {
		switch (CONNECT) {
		case MINMAX:
			result = Math.max(x, y);
			break;
		case LIMITED:
			result = Math.min(1, x + y);
			break;
		case ALGEBRAIC:
			result = (x + y) - (x * y);
			break;
		case TM:
			result = Math.max(x, y);
			break;
		case DRASTIC:
			if (Math.min(x, y) == 0)
				result = Math.max(x, y);
			else
				result = 1;
		}
		// if (result<alphaCut) result=0;
		return result;
	}

	public static final double snorm(double x, double y) {
		double result = 0;
		switch (CONNECT) {
		case MINMAX:
			result = Math.max(x, y);
			break;
		case LIMITED:
			result = Math.min(1, x + y);
			break;
		case ALGEBRAIC:
			result = (x + y) - (x * y);
			break;
		case TM:
			result = Math.max(x, y);
			break;
		case DRASTIC:
			if (Math.min(x, y) == 0)
				result = Math.max(x, y);
			else
				result = 1;
		}
		// if (result<alphaCut) result=0;
		return result;
	}

	// composition for FDM
	public static final float compose2(float x, float y) {
		return (float) compose((double) x, (double) y);
	}

	// composition for FDM
	public static final double compose2(double x, double y) {
		double result = 0;
		switch (COMPOSE) {
		case MINMAX:
			result = Math.max(x, y);
			break;
		case LIMITED:
			result = Math.min(1, x + y);
			break;
		case ALGEBRAIC:
			result = (x + y) - (x * y);
			break;
		case TM:
			result = Math.max(x, y);
			break;
		case DRASTIC:
			if (Math.min(x, y) == 0)
				result = Math.max(x, y);
			else
				result = 1;
		}
		// if (result<alphaCut) result=0;
		return result;
	}

	public final float getMembership(double object) {
		return membership((float) object);
	}

	public final float getMembership(float object) {
		return membership(object);
	}

	public final float getMembership(long object) {
		return membership(object);
	}

	public final float membership(float index) {
		float mem = 0;

		switch (TYPE) {
		case TRIANGULAR:
			mem = triangular(index);
			break;

		case TRAPEZOIDAL:
			mem = trapezoidal(index);
			break;

		case L_SHOULDERED:
			mem = leftShouldered(index);
			break;

		case R_SHOULDERED:
			mem = rightShouldered(index);
			break;

		case GAUSSIAN:
			mem = gaussian(index);
			break;

		case INDIVIDUAL:
			mem = individual(index);
			break;
		}
		return mem;
	}

	public final float triangular(float x) {
		float mem = 0;
		// mem = Math.max(Math.min((x-a)/(b-a),(c-x)/(c-b)),0);
		m1 = (x - a) / (b - a);
		m2 = (c - x) / (c - b);
		if (m1 < m2)
			mem = m1;
		else
			mem = m2;
		if (mem < 0)
			mem = 0;
		return mem;
	}

	public final float trapezoidal(float x) {
		float mem = 0;
		mem = Math.max(Math.min(Math.min((x - a) / (b - a), 1), (d - x)
				/ (d - c)), 0);
		return mem;
	}

	public final float leftShouldered(float x) {
		float mem = 0;
		if (x <= a)
			mem = 1;
		else if ((a < x) && (x < b))
			mem = 1 - ((x - a) / (b - a));
		else if (x >= b)
			mem = 0f;
		return mem;
	}

	public final float rightShouldered(float x) {
		float mem = 0;
		if (x <= a)
			mem = 0;
		else if ((a < x) && (x < b))
			mem = ((x - a) / (b - a));
		else if (x >= b)
			mem = 1;
		return mem;
	}

	// parameter a is mean, b is variance
	public final float gaussian(float x) {
		// d == sigma
		double val = -0.5 * ((a - x) * (a - x)) / (b);
		return (float) Math.exp(val);
	}

	public final float individual(float x) {
		return a;
	}

	public final FuzzySet clone1() {
		return new FuzzySet(name, i, a, b, c, d, TYPE);
	}

	public final String toString() {
		String ret = "";

		if (discrete) {
			// for (int i=0;i<values.size();i++)
			// ret+=String.valueOf(values.get(i))+",";
		} else {
			ret = name;
		}
		return ret;
	}

	public boolean separate(FuzzySet fs) {
		return ((TYPE == L_SHOULDERED && fs.TYPE == R_SHOULDERED) || (TYPE == R_SHOULDERED && fs.TYPE == L_SHOULDERED));
	}

	public static void and(FuzzySet fs) {
		/*
		 * switch(fs.TYPE) { case TRIANGULAR: switch(TYPE) { case TRIANGULAR: if
		 * ((a==fs.c)||(fs.a==c)) { a=b=c=0; empty=true; } else if (a<fs.c) {}
		 * break; } break; case LEFT_SHOULDERED: break; }
		 */
	}
}
